from pathlib import Path
import tempfile
import pickle

import pytest
import numpy as np
import matplotlib.pyplot as plt  # needed for visual inspection
import matplotlib.patches as mPatches
from shapely.geometry import Polygon as shPolygon
from shapely.geometry import MultiPolygon as shMultiPolygon

from locan import Region, Region1D, Region2D, Region3D, RegionND, EmptyRegion, Interval, \
    Rectangle, Ellipse, Polygon, MultiPolygon, \
    AxisOrientedCuboid, Cuboid, \
    AxisOrientedHypercuboid


def test_Region():
    with pytest.raises(TypeError):
        Region()

    region = Region.from_intervals((0, 2))
    assert repr(region) == 'Interval(0, 2)'
    region = Region.from_intervals(((0, 2), (0, 1)))
    assert repr(region) == 'Rectangle((0, 0), 2, 1, 0)'
    region = Region.from_intervals(((0, 1), (0, 2), (0, 3)))
    assert repr(region) ==  'AxisOrientedCuboid((0, 0, 0), 1, 2, 3)'
    region = Region.from_intervals(((0, 1), (0, 2), (0, 3), (0, 4)))
    assert repr(region) == 'AxisOrientedHypercuboid((0, 0, 0, 0), (1, 2, 3, 4))'


def test_Region1D():
    with pytest.raises(TypeError):
        Region1D()


def test_Region2D():
    with pytest.raises(TypeError):
        Region2D()
    points = ((2, 2), (2, 3), (3, 3), (3, 2.5), (2, 2))
    shapely_object = shPolygon(points)
    region = Region2D.from_shapely(shapely_object)
    assert isinstance(region, Polygon)

    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    shapely_object = shMultiPolygon([shapely_object, shPolygon(points, holes)])
    region = Region2D.from_shapely(shapely_object)
    assert isinstance(region, MultiPolygon)


def test_Region3D():
    with pytest.raises(TypeError):
        Region3D()


def test_RegionND():
    with pytest.raises(TypeError):
        RegionND()


def test_EmptyRegion():
    region = EmptyRegion()
    assert isinstance(region, Region)
    assert repr(region) == 'EmptyRegion()'
    assert str(region) == 'EmptyRegion()'
    assert region.dimension is None
    assert len(region.points) == 0
    assert region.centroid is None
    assert region.max_distance == 0
    assert region.region_measure == 0
    assert region.subregion_measure == 0
    assert region.bounds is None
    assert region.extent is None
    assert region.bounding_box is None
    other = Rectangle()
    assert isinstance(region.intersection(other), EmptyRegion)
    assert region.symmetric_difference(other) is other
    assert region.union(other) is other
    assert region.contains([[9, 8], [10.5, 10.5], [100, 100], [11, 12]]).size == 0
    assert region.contains([(10.5, 10)]).size == 0
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    with pytest.raises(NotImplementedError):
        region.as_artist()
    assert isinstance(region.shapely_object, shPolygon)
    with pytest.raises(NotImplementedError):
        region.buffer(1)


def test_Interval():
    region = Interval()
    assert not (2 in region)
    assert 0.5 in region
    assert isinstance(region, Region)
    assert repr(region) == 'Interval(0, 1)'
    assert str(region) == 'Interval(0, 1)'
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Interval)
    with pytest.raises(AttributeError):
        region.lower_bound = None
        region.upper_bound = None
    assert region.dimension == 1
    assert region.bounds == (0, 1)
    assert region.intervals == ((0, 1),)
    assert region.extent == 1
    assert np.allclose(region.points.astype(float), (0, 1))
    assert region.centroid == 0.5
    assert region.max_distance == 1
    assert region.region_measure == 1
    assert region.subregion_measure is None
    assert np.array_equal(region.contains((0, 0.5, 1, 2)), (0, 1))
    with pytest.raises(NotImplementedError):
        other = Interval()
        assert isinstance(region.intersection(other), EmptyRegion)
        assert region.symmetric_difference(other) is other
        assert region.union(other) is other
    assert region.contains((0.5,)) == (0,)
    assert region.contains((2,)).size == 0
    assert region.contains([]).size == 0
    # assert isinstance(region.as_artist(), mPatches)
    assert np.array_equal(region.buffer(1).points, [-1, 2])
    assert repr(region.buffer(1)) == 'Interval(-1, 2)'

    region = Interval.from_intervals((0, 2))
    assert repr(region) == 'Interval(0, 2)'
    region = Interval.from_intervals([0, 2])
    assert repr(region) == 'Interval(0, 2)'
    region = Interval.from_intervals(np.array([0, 2]))
    assert repr(region) == 'Interval(0, 2)'


def test_Rectangle():
    region = Rectangle((0, 0), 2, 1, 90)
    assert not ((10, 1) in region)
    assert (-0.5, 0.5) in region
    assert isinstance(region, Region)
    assert repr(region) == 'Rectangle((0, 0), 2, 1, 90)'
    assert str(region) == 'Rectangle((0, 0), 2, 1, 90)'
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Rectangle)
    with pytest.raises(AttributeError):
        region.corner = None
        region.width = None
        region.height = None
        region.angle = None
    assert region.dimension == 2
    assert region.bounds == pytest.approx((-1, 0, 0, 2))
    assert region.intervals == (pytest.approx((-1, 0)), pytest.approx((0, 2)))
    assert region.extent == pytest.approx((1, 2))
    assert len(region.points) == 5
    assert np.allclose(region.points.astype(float), [[0., 0.], [-1., 0.], [-1., 2.], [0., 2.], [0., 0.]])
    assert region.centroid == (-0.5, 1)
    assert region.max_distance == np.sqrt(5)
    assert region.region_measure == 2
    assert region.subregion_measure == 6
    assert np.array_equal(region.contains([[0, 0], [-0.5, 0.5], [100, 100], [-1, 2]]), (1,))
    other = Rectangle((0, 0), 2, 1, 0)
    assert isinstance(region.intersection(other), Polygon)
    assert isinstance(region.symmetric_difference(other), MultiPolygon)
    assert isinstance(region.union(other), Polygon)

    assert region.contains([(-0.5, 1)]) == (0,)
    assert region.contains([(10, 10)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.Rectangle)
    assert isinstance(region.shapely_object, shPolygon)
    assert region.region_measure == region.shapely_object.area
    assert isinstance(region.buffer(1), Polygon)
    assert region.bounding_box.corner == (-1., 0.)
    assert region.bounding_box.width == pytest.approx(1)
    assert region.bounding_box.height == pytest.approx(2)

    region = Rectangle.from_intervals(((0, 2), (0, 1)))
    assert repr(region) == 'Rectangle((0, 0), 2, 1, 0)'
    region = Rectangle.from_intervals([(0, 2), (0, 1)])
    assert repr(region) == 'Rectangle((0, 0), 2, 1, 0)'
    region = Rectangle.from_intervals(np.array([(0, 2), (0, 1)]))
    assert repr(region) == 'Rectangle((0, 0), 2, 1, 0)'


def test_Rectangle_visual():
    region = Rectangle((0, 0), 2, 1, 90)
    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*region.points.T, marker='o', color='Blue')
    ax.add_patch(region.as_artist(origin=(0, 0), fill=True, alpha=0.2))
    ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker='.', color='Red')
    ax.plot(*region.centroid, '*', color='Green')
    ax.plot(*np.array(region.buffer(1).exterior.coords).T, marker='.', color='Yellow')
    region.plot(color='Green', alpha=0.2)
    # plt.show()
    plt.close('all')


def test_Ellipse():
    region = Ellipse((10, 10), 4, 2, 90)
    assert not ((1, 1) in region)
    assert (10, 10) in region
    assert isinstance(region, Region)
    assert repr(region) == 'Ellipse((10, 10), 4, 2, 90)'
    assert str(region) == 'Ellipse((10, 10), 4, 2, 90)'
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Ellipse)
    with pytest.raises(AttributeError):
        region.center = None
        region.width = None
        region.height = None
        region.angle = None
    assert region.dimension == 2
    assert region.bounds == pytest.approx((9, 8, 11, 12))
    assert region.extent == pytest.approx((2, 4))
    assert (len(region.points) == 65 or len(region.points) == 66)
    assert region.centroid == (10, 10)
    assert region.max_distance == 4
    assert region.region_measure == pytest.approx(6.283185307179586)
    assert region.subregion_measure == pytest.approx(9.688448216130086)
    assert np.array_equal(region.contains([[9, 8], [10.5, 10.5], [100, 100], [11, 12]]), (1,))
    other = Rectangle((10, 10), 10, 10, 0)
    assert isinstance(region.intersection(other), Polygon)
    assert isinstance(region.symmetric_difference(other), MultiPolygon)
    assert isinstance(region.union(other), Polygon)
    assert region.contains([(10.5, 10)]) == (0,)
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.Ellipse)
    assert isinstance(region.shapely_object, shPolygon)
    assert region.region_measure == pytest.approx(region.shapely_object.area, rel=10e-3)
    assert isinstance(region.buffer(1), Polygon)
    assert region.bounding_box.corner == (9, 8)
    assert region.bounding_box.width == pytest.approx(2)
    assert region.bounding_box.height == pytest.approx(4)


def test_Ellipse_visual():
    region = Ellipse((10, 10), 4, 2, 90)
    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*region.points.T, marker='o', color='Blue')
    ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker='.', color='Red')
    ax.add_patch(region.as_artist(origin=(0, 0), fill=True, alpha=0.2))
    ax.plot(*np.array(region.buffer(1).exterior.coords).T, marker='.', color='Yellow')
    ax.plot(*region.centroid, '*', color='Green')
    region.plot(color='Green', alpha=0.2)
    # plt.show()
    plt.close('all')


def test_Polygon():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    region = Polygon(points)
    assert not ((10, 1) in region)
    assert (0.5, 0.5) in region
    assert np.array_equal(region.points, points)
    assert repr(region) == 'Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]])'
    region = Polygon(points[:-1])
    assert isinstance(region, Region)
    assert repr(region) == 'Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]])'
    assert str(region) == 'Polygon(<self.points>, <self.holes>)'
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Polygon)
    assert region.dimension == 2
    assert np.array_equal(region.points, points)
    assert region.centroid == (pytest.approx(0.4444444444444444), pytest.approx(0.611111111111111))
    assert region.max_distance == np.sqrt(2)
    assert region.region_measure == pytest.approx(0.75)
    assert region.subregion_measure == pytest.approx(3.618033988749895)

    assert np.array_equal(region.contains([[0, 0], [0.2, 0.8], [100, 100], [1, 0.5]]), (1,))
    assert region.contains([(0.2, 0.8)]) == (0,)
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.PathPatch)
    assert isinstance(region.shapely_object, shPolygon)
    assert region.region_measure == pytest.approx(region.shapely_object.area, rel=10e-3)
    assert isinstance(region.buffer(1), Polygon)
    assert region.bounding_box.corner == (0, 0)
    assert region.bounding_box.width == pytest.approx(1)
    assert region.bounding_box.height == pytest.approx(1)


def test_Polygon_visual():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    region = Polygon(points)
    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*region.points.T, marker='.', color='Blue')
    ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker='.', color='Red')
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    ax.plot(*np.array(region.buffer(1).exterior.coords).T, marker='.', color='Yellow')
    ax.plot(*region.centroid, '*', color='Green')
    region.plot(color='Green', alpha=0.2)
    # plt.show()

    # visualize points inside
    # points = np.random.default_rng().random(size=(10, 2))
    # points_inside = points[region.contains(points)]
    # fig, ax = plt.subplots(nrows=1, ncols=1)
    # ax.scatter(*points_inside.T, marker='.', color='Blue')
    # ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker='.', color='Red')
    # ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    # plt.show()

    plt.close('all')


def test_Polygon_with_holes():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region = Polygon(points, holes)
    assert np.array_equal(region.points, points)
    assert repr(region) == 'Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]], ' \
                           '[[[0.2, 0.2], [0.2, 0.3], [0.3, 0.3], [0.3, 0.25]], ' \
                           '[[0.4, 0.4], [0.4, 0.5], [0.5, 0.5], [0.5, 0.45]]])'
    region = Polygon(points[:-1], holes)
    assert isinstance(region, Region)
    assert repr(region) == 'Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]], ' \
                           '[[[0.2, 0.2], [0.2, 0.3], [0.3, 0.3], [0.3, 0.25]], ' \
                           '[[0.4, 0.4], [0.4, 0.5], [0.5, 0.5], [0.5, 0.45]]])'
    assert str(region) == 'Polygon(<self.points>, <self.holes>)'
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Polygon)
    assert region.dimension == 2
    assert np.array_equal(region.points, points)
    assert region.centroid == (pytest.approx(0.446485260770975), pytest.approx(0.6162131519274376))
    assert region.max_distance == np.sqrt(2)
    assert region.region_measure == pytest.approx(0.735)
    assert region.subregion_measure == pytest.approx(4.341640786499874)

    assert np.array_equal(region.contains([[0, 0], [0.2, 0.8], [100, 100], [1, 0.5]]), (1,))
    assert region.contains([(0.2, 0.8)]) == (0,)
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.PathPatch)
    assert isinstance(region.shapely_object, shPolygon)
    assert region.region_measure == pytest.approx(region.shapely_object.area, rel=10e-3)
    assert isinstance(region.buffer(1), Polygon)
    assert region.bounding_box.corner == (0, 0)
    assert region.bounding_box.width == pytest.approx(1)
    assert region.bounding_box.height == pytest.approx(1)


def test_Polygon_with_holes_visual():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region = Polygon(points, holes)
    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*region.points.T, marker='o', color='Blue')
    ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker='.', color='Red')
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    ax.plot(*np.array(region.buffer(0.01).exterior.coords).T, marker='.', color='Yellow')
    ax.plot(*region.centroid, '*', color='Green')
    region.plot(color='Green', alpha=0.2)
    #plt.show()

    plt.close('all')


def test_Polygon_from_shapely():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    shapely_polygon = shPolygon(points, holes)
    region = Polygon.from_shapely(shapely_polygon)
    assert np.array_equal(region.points, points)
    assert isinstance(region, Region)
    assert repr(region) == 'Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]], ' \
                           '[[[0.2, 0.2], [0.2, 0.3], [0.3, 0.3], [0.3, 0.25], [0.2, 0.2]], ' \
                           '[[0.4, 0.4], [0.4, 0.5], [0.5, 0.5], [0.5, 0.45], [0.4, 0.4]]])'
    assert str(region) == 'Polygon(<self.points>, <self.holes>)'
    assert region.dimension == 2
    assert np.array_equal(region.points, points)
    assert region.centroid == (pytest.approx(0.446485260770975), pytest.approx(0.6162131519274376))
    assert region.max_distance == np.sqrt(2)
    assert region.region_measure == pytest.approx(0.735)
    assert region.subregion_measure == pytest.approx(4.341640786499874)

    assert np.array_equal(region.contains([[0, 0], [0.2, 0.8], [100, 100], [1, 0.5]]), (1,))
    assert region.contains([(0.2, 0.8)]) == (0,)
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.PathPatch)
    assert isinstance(region.shapely_object, shPolygon)
    assert region.region_measure == pytest.approx(region.shapely_object.area, rel=10e-3)
    assert isinstance(region.buffer(1), Polygon)


@pytest.mark.visual
def test_Polygon_from_shapely_visual():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    shapely_polygon = shPolygon(points, holes)
    region = Polygon.from_shapely(shapely_polygon)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*region.points.T, marker='o', color='Blue')
    ax.plot(*np.array(region.shapely_object.exterior.coords).T, marker='.', color='Red')
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    ax.plot(*np.array(region.buffer(0.01).exterior.coords).T, marker='.', color='Yellow')
    ax.plot(*region.centroid, '*', color='Green')
    region.plot(color='Green', alpha=0.2)
    plt.show()

    plt.close('all')


def test_MultiPolygon():
    points = ((2, 2), (2, 3), (3, 3), (3, 2.5), (2, 2))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region_1 = Polygon(points, holes)
    region = MultiPolygon([region_0, region_1])
    assert not ((10, 1) in region)
    assert (0.2, 0.8) in region
    assert isinstance(region, Region)
    assert repr(region) == 'MultiPolygon([Polygon([[2.0, 2.0], [2.0, 3.0], [3.0, 3.0], [3.0, 2.5], [2.0, 2.0]]), Polygon([[0.0, 0.0], [0.0, 1.0], [1.0, 1.0], [1.0, 0.5], [0.0, 0.0]], [[[0.2, 0.2], [0.2, 0.3], [0.3, 0.3], [0.3, 0.25]], [[0.4, 0.4], [0.4, 0.5], [0.5, 0.5], [0.5, 0.45]]])])'
    assert str(region) == 'MultiPolygon(<self.polygons>)'
    new_reg = eval(repr(region))
    assert isinstance(new_reg, MultiPolygon)
    assert region.dimension == 2
    assert len(region.polygons) == 2
    assert len(region.points) == 2
    assert len(region.holes) == 2
    assert region.centroid == (pytest.approx(1.4555555555555553), pytest.approx(1.6237373737373733))
    assert region.max_distance == pytest.approx(4.242640687119285)
    assert region.region_measure == pytest.approx(1.485)
    assert region.subregion_measure == pytest.approx(7.959674775249769)
    assert np.array_equal(region.contains([[0, 0], [0.2, 0.8], [100, 100], [1, 0.5]]), (1,))
    assert region.contains([(0.2, 0.8)]) == (0,)
    assert region.contains([(100, 100)]).size == 0
    assert region.contains([]).size == 0
    assert isinstance(region.as_artist(), mPatches.PathPatch)
    assert isinstance(region.shapely_object, shMultiPolygon)
    assert region.region_measure == pytest.approx(region.shapely_object.area, rel=10e-3)
    assert isinstance(region.buffer(1), Polygon)
    assert region.bounds == (0, 0, 3, 3)
    assert region.extent == (3, 3)
    assert region.bounding_box.corner == (0, 0)
    assert region.bounding_box.width == pytest.approx(3)
    assert region.bounding_box.height == pytest.approx(3)


def test_MultiPolygon_visual():
    points = ((2, 2), (2, 3), (3, 3), (3, 2.5), (2, 2))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region_1 = Polygon(points, holes)
    region = MultiPolygon([region_0, region_1])

    fig, ax = plt.subplots(nrows=1, ncols=1)
    for polygon in region.polygons:
        ax.plot(*polygon.points.T, marker='o', color='Blue')
        ax.plot(*np.array(polygon.shapely_object.exterior.coords).T, marker='.', color='Red')
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    ax.plot(*np.array(region.buffer(1).exterior.coords).T, marker='.', color='Yellow')
    ax.plot(*region.centroid, '*', color='Green')
    region.plot(color='Green', alpha=0.2)
    # plt.show()

    # visualize points inside
    # points = np.random.default_rng().random(size=(100, 2)) * 5
    # points_inside = points[region.contains(points)]
    # fig, ax = plt.subplots(nrows=1, ncols=1)
    # ax.plot(*points.T, '.', markersize=1, color='Gray')
    # ax.plot(*points_inside.T, 'o', markersize=1, color='Blue')
    # for polygon in region.polygons:
    #     ax.plot(*np.array(polygon.shapely_object.exterior.coords).T, marker='.', color='Red')
    # ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    # plt.show()

    plt.close('all')


def test_Polygon_union():
    points = ((2, 2), (2, 3), (3, 3), (3, 2.5), (2, 2))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region_1 = Polygon(points, holes)
    region = region_0.union(region_1)
    assert isinstance(region, Region)
    assert isinstance(region, Region2D)
    assert isinstance(region, MultiPolygon)
    assert region.dimension == 2
    assert len(region.polygons) == 2
    assert len(region.points) == 2
    assert len(region.holes) == 2


@pytest.mark.visual
def test_Polygon_union_visual():
    points = ((2, 2), (2, 3), (3, 3), (3, 2.5), (2, 2))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region_1 = Polygon(points, holes)
    region = region_0.union(region_1)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.add_patch(region_0.as_artist(fill=True, alpha=0.2, color='Red'))
    ax.add_patch(region_1.as_artist(fill=True, alpha=0.2, color='Green'))
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.plot(*region.centroid, '*', color='Green')
    region.plot(color='Green', alpha=0.2)
    plt.show()

    # visualize points inside
    points = np.random.default_rng().random(size=(1_000, 2)) * 10
    points_inside = points[region.contains(points)]
    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*points_inside.T, '.', markersize=1, color='Blue')
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    plt.show()

    plt.close('all')


def test_Polygon_union_ragged():
    points = ((0, 0), (0, 10), (10, 10), (10, 0))
    holes = [((1, 1), (1, 9), (4, 9), (4, 1))]
    region_0 = Polygon(points, holes)
    points = ((2, 2), (2, 5), (3, 5), (3, 2))
    holes = [((2.5, 2.5), (2.5, 4), (2.6, 4), (2.6, 2.5))]
    region_1 = Polygon(points, holes)
    region = region_0.union(region_1)
    assert region.region_measure == pytest.approx(region_0.region_measure + region_1.region_measure)


@pytest.mark.visual
def test_Polygon_union_ragged_visual():
    points = ((0, 0), (0, 10), (10, 10), (10, 0))
    holes = [((1, 1), (1, 9), (4, 9), (4, 1))]
    region_0 = Polygon(points, holes)
    points = ((2, 2), (2, 5), (3, 5), (3, 2))
    holes = [((2.5, 2.5), (2.5, 4), (2.6, 4), (2.6, 2.5))]
    region_1 = Polygon(points, holes)
    region = region_0.union(region_1)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.add_patch(region_0.as_artist(fill=True, alpha=0.2, color='Red'))
    ax.add_patch(region_1.as_artist(fill=True, alpha=0.2, color='Green'))
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.plot(*region.centroid, '*', color='Green')
    region.plot(color='Green', alpha=0.2)
    plt.show()

    # visualize points inside
    points = np.random.default_rng().random(size=(1_000, 2)) * 10
    points_inside = points[region.contains(points)]
    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*points_inside.T, '.', markersize=1, color='Blue')
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    plt.show()

    plt.close('all')


def test_Polygon_intersection():
    points = ((0, 2), (0, 3), (3, 3), (3, 2.5), (2, 2))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region_1 = Polygon(points, holes)
    region = region_0.intersection(region_1)
    assert isinstance(region, EmptyRegion)

    points = ((0, 0), (0, 3), (3, 3), (3, 2), (2, 1.5))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region_1 = Polygon(points, holes)
    region = region_0.intersection(region_1)
    assert isinstance(region, Polygon)
    assert region.dimension == 2
    assert len(region.points) == 5
    assert len(region.holes) == 2


@pytest.mark.visual
def test_Polygon_intersection_visual():
    points = ((0, 0), (0, 3), (3, 3), (3, 2), (2, 1.5))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    region_1 = Polygon(points, holes)
    region = region_0.intersection(region_1)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.add_patch(region_0.as_artist(fill=True, alpha=0.2, color='Red'))
    ax.add_patch(region_1.as_artist(fill=True, alpha=0.2, color='Green'))
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.plot(*region.centroid, '*', color='Green')
    region.plot(color='Green', alpha=0.2)
    plt.show()

    # visualize points inside
    points = np.random.default_rng().random(size=(100_000, 2)) * 10
    points_inside = points[region.contains(points)]
    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*points_inside.T, '.', markersize=1, color='Blue')
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    plt.show()

    plt.close('all')


def test_Polygon_difference():
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.4), (0.4, 0.4), (0.3, 0.25)), ((0.5, 0.5), (0.5, 0.8), (0.8, 0.8), (0.7, 0.45))]
    region_0 = Polygon(points, holes)
    other_region = Rectangle(corner=(0.5, 0.2), width=0.5, height=0.5, angle=0)
    region = region_0.symmetric_difference(other_region)

    points = ((0, 0), (0, 3), (3, 3), (3, 2), (2, 1.5))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    other_region = Polygon(points, holes)
    region = region_0.symmetric_difference(other_region)
    # print(repr(region))
    assert isinstance(region, MultiPolygon)
    assert region.dimension == 2
    assert len(region.polygons) == 4


@pytest.mark.visual
def test_Polygon_difference_visual():
    points = ((0, 0), (0, 3), (3, 3), (3, 2), (2, 1.5))
    region_0 = Polygon(points)
    points = ((0, 0), (0, 1), (1, 1), (1, 0.5), (0, 0))
    holes = [((0.2, 0.2), (0.2, 0.3), (0.3, 0.3), (0.3, 0.25)), ((0.4, 0.4), (0.4, 0.5), (0.5, 0.5), (0.5, 0.45))]
    other_region = Polygon(points, holes)
    region = region_0.symmetric_difference(other_region)

    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.add_patch(region_0.as_artist(fill=True, alpha=0.2, color='Red'))
    ax.add_patch(other_region.as_artist(fill=True, alpha=0.2, color='Green'))
    ax.add_patch(region.as_artist(fill=True, alpha=0.2, color='Blue'))
    ax.plot(*region.centroid, '*', color='Green')
    region.plot(color='Green', alpha=0.2)
    plt.show()

    # visualize points inside
    points = np.random.default_rng().random(size=(100_000, 2)) * 10
    points_inside = points[region.contains(points)]
    fig, ax = plt.subplots(nrows=1, ncols=1)
    ax.plot(*points_inside.T, '.', markersize=1, color='Blue')
    ax.add_patch(region.as_artist(fill=True, alpha=0.2))
    plt.show()

    plt.close('all')


def test_AxisOrientedCuboid():
    region = AxisOrientedCuboid((1, 1, 1), 9, 19, 29)
    assert not ((-1, 1, 1) in region)
    assert (2, 2, 2) in region
    assert isinstance(region, Region)
    assert isinstance(region, Region3D)
    assert repr(region) == 'AxisOrientedCuboid((1, 1, 1), 9, 19, 29)'
    assert str(region) == 'AxisOrientedCuboid((1, 1, 1), 9, 19, 29)'
    new_reg = eval(repr(region))
    assert isinstance(new_reg, AxisOrientedCuboid)
    with pytest.raises(AttributeError):
        region.corner = None
        region.length = None
        region.width = None
        region.height = None
    assert region.dimension == 3
    assert region.bounds == pytest.approx((1, 1, 1, 10, 20, 30))
    assert region.intervals == (pytest.approx((1, 10)), pytest.approx((1, 20)), pytest.approx((1, 30)))
    assert region.extent == pytest.approx((9, 19, 29))
    assert len(region.points) == 8
    assert region.centroid == (5.5, 10.5, 15.5)
    assert region.max_distance == np.sqrt(9**2+19**2+29**2)
    assert region.region_measure == (9 * 19 * 29)
    assert region.subregion_measure == 2 * (9 * 19 + 19 * 29 + 29 * 9)
    assert np.array_equal(region.contains([[0, 0, 0], [1, 10, 10], [10, 10, 10], [5, 100, 10], [5, 10, 100]]), (1,))
    #other = Rectangle((0, 0), 2, 1, 0)
    #assert isinstance(region.intersection(other), Polygon)
    #assert isinstance(region.symmetric_difference(other), MultiPolygon)
    #assert isinstance(region.union(other), Polygon)

    assert region.contains([(2, 2, 2)]) == (0,)
    assert region.contains([(0, 0, 0)]).size == 0
    assert region.contains([]).size == 0
    # assert isinstance(region.as_artist(), mPatches.Rectangle)
    assert isinstance(region.buffer(1), AxisOrientedCuboid)
    assert repr(region.buffer(1)) == 'AxisOrientedCuboid((0, 0, 0), 11, 21, 31)'
    assert region.bounding_box.corner == (1, 1, 1)
    assert region.bounding_box.length == pytest.approx(9)
    assert region.bounding_box.width == pytest.approx(19)
    assert region.bounding_box.height == pytest.approx(29)

    region = AxisOrientedCuboid.from_intervals(((1, 10), (1, 20), (1, 30)))
    assert repr(region) == 'AxisOrientedCuboid((1, 1, 1), 9, 19, 29)'
    region = AxisOrientedCuboid.from_intervals([(1, 10), (1, 20), (1, 30)])
    assert repr(region) == 'AxisOrientedCuboid((1, 1, 1), 9, 19, 29)'
    region = AxisOrientedCuboid.from_intervals(np.array([(1, 10), (1, 20), (1, 30)]))
    assert repr(region) == 'AxisOrientedCuboid((1, 1, 1), 9, 19, 29)'


def test_Cuboid():
    region = Cuboid((1, 1, 1), 9, 19, 29, 0, 0, 0)
    assert isinstance(region, Region)
    assert isinstance(region, Region3D)
    assert repr(region) == 'Cuboid((1, 1, 1), 9, 19, 29, 0, 0, 0)'
    assert str(region) == 'Cuboid((1, 1, 1), 9, 19, 29, 0, 0, 0)'
    new_reg = eval(repr(region))
    assert isinstance(new_reg, Cuboid)
    with pytest.raises(AttributeError):
        region.corner = None
        region.length = None
        region.width = None
        region.height = None
        region.alpha = None
        region.beta = None
        region.gamma = None
    assert region.dimension == 3
    # assert region.bounds == pytest.approx((1, 1, 1, 10, 20, 30))
    # assert region.extent == pytest.approx((9, 19, 29))
    # assert len(region.points) == 8
    # assert region.centroid == (5.5, 10.5, 15.5)
    assert region.max_distance == np.sqrt(9**2+19**2+29**2)
    assert region.region_measure == (9 * 19 * 29)
    assert region.subregion_measure == 2 * (9 * 19 + 19 * 29 + 29 * 9)
    # assert np.array_equal(region.contains([[0, 0, 0], [1, 10, 10], [10, 10, 10], [5, 100, 10], [5, 10, 100]]), (1,))
    #other = Rectangle((0, 0), 2, 1, 0)
    #assert isinstance(region.intersection(other), Polygon)
    #assert isinstance(region.symmetric_difference(other), MultiPolygon)
    #assert isinstance(region.union(other), Polygon)

    # assert region.contains([(2, 2, 2)]) == (0,)
    # assert region.contains([(0, 0, 0)]).size == 0
    # assert region.contains([]).size == 0
    # assert isinstance(region.as_artist(), mPatches.Rectangle)
    # assert isinstance(region.buffer(1), AxisOrientedCuboid)
    # assert repr(region.buffer(1)) == 'AxisOrientedCuboid((0, 0, 0), 11, 21, 31)'
    # assert region.bounding_box.corner == (1, 1, 1)
    # assert region.bounding_box.length == pytest.approx(9)
    # assert region.bounding_box.width == pytest.approx(19)
    # assert region.bounding_box.height == pytest.approx(29)


def test_AxisOrientedHypercuboid():
    region = AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))
    assert not ((-1, 1, 1) in region)
    assert (2, 2, 2) in region
    assert isinstance(region, Region)
    assert repr(region) == 'AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))'
    assert str(region) == 'AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))'
    new_reg = eval(repr(region))
    assert isinstance(new_reg, AxisOrientedHypercuboid)
    with pytest.raises(AttributeError):
        region.corner = None
        region.length = None
        region.width = None
        region.height = None
    assert region.dimension == 3
    assert region.bounds == pytest.approx((1, 1, 1, 10, 20, 30))
    assert region.intervals == (pytest.approx((1, 10)), pytest.approx((1, 20)), pytest.approx((1, 30)))
    assert region.extent == pytest.approx((9, 19, 29))
    assert len(region.points) == 8
    assert region.centroid.tolist() == [5.5, 10.5, 15.5]
    assert region.max_distance == np.sqrt(9**2+19**2+29**2)
    assert region.region_measure == (9 * 19 * 29)
    with pytest.raises(NotImplementedError):
        region.subregion_measure
    assert np.array_equal(region.contains([[0, 0, 0], [1, 10, 10], [10, 10, 10], [5, 100, 10], [5, 10, 100]]), (1,))
    # needs to be implemented:
    #other = Rectangle((0, 0), 2, 1, 0)
    #assert isinstance(region.intersection(other), Polygon)
    #assert isinstance(region.symmetric_difference(other), MultiPolygon)
    #assert isinstance(region.union(other), Polygon)

    assert region.contains([(2, 2, 2)]) == (0,)
    assert region.contains([(0, 0, 0)]).size == 0
    assert region.contains([]).size == 0
    # assert isinstance(region.as_artist(), mPatches.Rectangle)
    assert isinstance(region.buffer(1), AxisOrientedHypercuboid)
    assert repr(region.buffer(1)) == 'AxisOrientedHypercuboid((0, 0, 0), (11, 21, 31))'
    assert region.bounding_box.corner.tolist() == [1, 1, 1]
    assert region.bounding_box.lengths.tolist() == [9, 19, 29]

    region = AxisOrientedHypercuboid.from_intervals(((1, 10), (1, 20), (1, 30)))
    assert repr(region) == 'AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))'
    region = AxisOrientedHypercuboid.from_intervals([(1, 10), (1, 20), (1, 30)])
    assert repr(region) == 'AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))'
    region = AxisOrientedHypercuboid.from_intervals(np.array([(1, 10), (1, 20), (1, 30)]))
    assert repr(region) == 'AxisOrientedHypercuboid((1, 1, 1), (9, 19, 29))'


def test_pickling_Region():
    regions = [EmptyRegion(), Interval(), Rectangle(), Ellipse(), Polygon(((0, 0), (0, 1), (2, 2))),
               MultiPolygon([Polygon(((10, 10), (10, 11), (12, 12))), Polygon(((0, 0), (0, 1), (2, 2)))])]
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / 'pickled_region.pickle'
        for region in regions:
            with open(file_path, 'wb') as file:
                pickle.dump(region, file, pickle.HIGHEST_PROTOCOL)
            with open(file_path, 'rb') as file:
                region_pickled = pickle.load(file)
            assert type(region_pickled) == type(region)



from pathlib import Path

import pytest
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt  # this import is needed for visual inspection

from locan import LocData
from locan import ROOT_DIR
from locan.locan_io.locdata.io_locdata import load_txt_file
from locan.data.transform import transform_affine
from locan.analysis.cbc import _coordinate_based_colocalization
from locan.analysis import CoordinateBasedColocalization


# fixtures

@pytest.fixture()
def locdata():
    path = Path(ROOT_DIR / 'tests/test_data/five_blobs.txt')
    dat = load_txt_file(path)
    return dat


@pytest.fixture()
def locdata_3d():
    path = Path(ROOT_DIR / 'tests/test_data/five_blobs_3D.txt')
    dat = load_txt_file(path)
    return dat


@pytest.fixture()
def locdata_line():
    locdata_dict = {
        'position_x': [0, 1, 3, 4, 50, 98, 99, 100],
        'position_y': np.zeros(8)
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


def test_cbc(locdata):
    points = locdata.coordinates
    points_trans = transform_affine(locdata.coordinates)
    res = _coordinate_based_colocalization(points=points, other_points=points_trans, radius=100, n_steps=10)
    assert(res[0] == 0.6162215398589203)
    # print(res[0:5])

    # # plot data
    # fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(10, 5))
    # axes[0].scatter(x=points[:,0], y=points[:,1], marker='.', c=res, cmap='jet', label='points')
    # axes[0].set_title('CBC')
    #
    # axes[1].scatter(x=points[:,0], y=points[:,1], marker='.', color='Blue', label='points')
    # axes[1].scatter(x=points_trans[:,0], y=points_trans[:,1], marker='o', color='Red', label='transformed points')
    # axes[1].set_title('Transformed points')
    #
    # plt.show()

    plt.close('all')


def test_cbc_nan(locdata_line):
    res = _coordinate_based_colocalization(points=locdata_line.coordinates,
                                           other_points=locdata_line.coordinates+(0.5, 0),
                                           radius=10, n_steps=10)
    assert(len(res) == 8)
    assert(np.isnan(res[4]))


def test_Coordinate_based_colocalization_empty(caplog):
    cbc = CoordinateBasedColocalization().compute(LocData())
    cbc.hist()
    assert caplog.record_tuples == [('locan.analysis.cbc', 30, 'Locdata is empty.')]

    plt.close('all')


def test_Coordinate_based_colocalization(locdata_line):
    other_locdata = transform_affine(locdata_line)
    cbc = CoordinateBasedColocalization(radius=100, n_steps=10).compute(locdata=locdata_line,
                                                                        other_locdata=other_locdata)
    # print(cbc.results)
    assert(cbc.results.columns == f'colocalization_cbc_{other_locdata.meta.identifier}')

    cbc.hist()

    plt.close('all')



from platform import system
from io import StringIO, BytesIO, TextIOWrapper
import logging

import pytest
import numpy as np
from locan import ROOT_DIR
from locan.constants import DECODE_KEYS, THUNDERSTORM_KEYS
from locan.locan_io import convert_property_types, convert_property_names

from locan.locan_io.locdata.utilities import open_path_or_file_like


def test_convert_property_types(locdata_2d):
    df = locdata_2d.data.copy()
    types_mapping = {'position_x': 'float', 'position_y': str, 'frame': np.int64, 'not_in_there': 'float'}
    converted_df = convert_property_types(dataframe=df, loc_properties=None, types=types_mapping)
    assert converted_df.dtypes[0] == np.float32
    assert converted_df.dtypes[1] == object
    assert converted_df.dtypes[2] == np.int64
    assert converted_df.dtypes[3] == np.int32
    assert isinstance(converted_df['position_y'].iloc[0], str)


def test_open_path_or_file_like():
    if system() == 'Linux':
        inputs = [
            str(ROOT_DIR) + '/tests/test_data/rapidSTORM_dstorm_data.txt',
            ROOT_DIR / 'tests/test_data/rapidSTORM_dstorm_data.txt',
        ]
    elif system() == 'Windows':
        inputs = [
            str(ROOT_DIR) + '/tests/test_data/rapidSTORM_dstorm_data.txt',
            str(ROOT_DIR) + r'\tests\test_data\rapidSTORM_dstorm_data.txt',
            ROOT_DIR / 'tests/test_data/rapidSTORM_dstorm_data.txt',
            ROOT_DIR / r'tests\test_data\rapidSTORM_dstorm_data.txt'
        ]
    else:
        inputs = [
            str(ROOT_DIR) + '/tests/test_data/rapidSTORM_dstorm_data.txt',
            ROOT_DIR / 'tests/test_data/rapidSTORM_dstorm_data.txt',
        ]

    for pfl in inputs:
        with open_path_or_file_like(path_or_file_like=pfl) as data:
            assert isinstance(data, TextIOWrapper)
            out = data.read(10)
            assert out and isinstance(out, str)
            assert not data.closed
        assert data.closed

    pfl = StringIO("This is a file-like object to be read.")
    with open_path_or_file_like(path_or_file_like=pfl) as data:
        assert isinstance(data, StringIO)
        out = data.read(10)
        assert out and isinstance(out, str)
        assert not data.closed
    assert data.closed

    pfl = BytesIO(b"This is a file-like object to be read.")
    with open_path_or_file_like(path_or_file_like=pfl) as data:
        assert isinstance(data, BytesIO)
        out = data.read(10)
        assert out and isinstance(out, bytes)
        assert not data.closed
    assert data.closed

    pfl = ROOT_DIR / 'tests/test_data/some_file_that_does_not_exits.txt'
    with pytest.raises(FileNotFoundError):
        with open_path_or_file_like(path_or_file_like=pfl):
            pass


def test_convert_property_names(caplog):
    new_properties = convert_property_names(properties=['position_x'],
                                            property_mapping=None)
    assert new_properties == ['position_x']

    properties = ['position_x', 'y', 'z', 'frame_ix', 'something']
    new_properties = convert_property_names(properties=properties,
                                            property_mapping=None)
    assert new_properties == ['position_x', 'y', 'z', 'frame_ix', 'something']
    assert caplog.record_tuples[0] == ('locan.locan_io.locdata.utilities', logging.WARNING,
                                       'Column y is not a Locan property standard.')

    new_properties = convert_property_names(properties=properties,
                                            property_mapping=DECODE_KEYS)
    assert new_properties == ['position_x', 'position_y', 'position_z', 'frame', 'something']
    assert caplog.record_tuples[4] == ('locan.locan_io.locdata.utilities', logging.WARNING,
                                       'Column something is not a Locan property standard.')

    properties = ['position_x', 'y', 'z', 'frame_ix', 'something', 'sigma [nm]']
    new_properties = convert_property_names(properties=properties,
                                            property_mapping=[DECODE_KEYS, THUNDERSTORM_KEYS])
    assert new_properties == ['position_x', 'position_y', 'position_z', 'frame', 'something',
                              'psf_sigma']
    assert caplog.record_tuples[4] == ('locan.locan_io.locdata.utilities', logging.WARNING,
                                       'Column something is not a Locan property standard.')



"""

File input/output for localization data in SMLM files.

File specifications are provided at https://github.com/imodpasteur/smlm-file-format/blob/master/specification.md.

Code is adapted from https://github.com/imodpasteur/smlm-file-format/blob/master/implementations/Python/smlm_file.py.
(MIT license)
"""

import time
import logging
import zipfile
import json

import numpy as np
import pandas as pd
from google.protobuf import json_format

from locan.locan_io.locdata import manifest_pb2
from locan.data.locdata import LocData
import locan.constants
from locan.data import metadata_pb2
from locan.locan_io.locdata.utilities import convert_property_types, convert_property_names
from locan.utils.format import _time_string

__all__ = ['manifest_format_from_locdata', 'manifest_file_info_from_locdata', 'manifest_from_locdata',
           'save_SMLM', 'load_SMLM_manifest', 'load_SMLM_header', 'load_SMLM_file']

logger = logging.getLogger(__name__)


dtype2length = {'int8': 1, 'uint8': 1, 'int16': 2, 'uint16': 2,
                'int32': 4, 'uint32': 4, 'int64': 8, 'uint64': 8,
                'float32': 4, 'float64': 8}


def manifest_format_from_locdata(locdata):
    """
    Prepare a manifest["format"] protobuf from locdata.
    The manifest holds metadata for smlm files.

    Parameters
    ----------
    locdata : LocData
        The LocData object.

    Returns
    -------
    locan.locan_io.locdata.manifest_pb2.Format
        The manifest format
    """
    format = manifest_pb2.Format()

    # required by
    format.name = "smlm-table(binary)-0"
    format.type = manifest_pb2.Type.TABLE
    format.mode = manifest_pb2.Mode.BINARY
    format.columns = len(locdata.data.columns)
    format.headers.extend(locdata.data.columns)
    format.dtype.extend(
        [manifest_pb2.Dtype.Value(str(dt).upper())
         for dt in locdata.data.dtypes.values]
    )
    format.shape.extend([1] * len(locdata.data.columns))
    format.units.extend([""] * len(locdata.data.columns))  # todo: add correct units

    # optional by specifications
    format.description = "localization table generated from locan.LocData"

    return format


def manifest_file_info_from_locdata(locdata):
    """
    Prepare a manifest["file_info"] protobuf from locdata.
    The manifest holds metadata for smlm files.

    Parameters
    ----------
    locdata : LocData
        The LocData object.

    Returns
    -------
    locan.locan_io.locdata.manifest_pb2.FileInfo
        The manifest file information
    """
    file_info = manifest_pb2.FileInfo()

    # required by specifications
    file_info.name = "table-0.bin"
    file_info.type = manifest_pb2.TABLE
    file_info.format = "smlm-table(binary)-0"
    file_info.channel = "default"
    file_info.rows = len(locdata)
    for column in locdata.data.columns:
        file_info.offset[column] = 0

    # optional by specifications
    for column in locdata.data.columns:
        file_info.min[column] = locdata.data[column].min()
    for column in locdata.data.columns:
        file_info.max[column] = locdata.data[column].max()
    # file_info.exposure = 20

    return file_info


def manifest_from_locdata(locdata, return_json_string=False):
    """
    Prepare a manifest protobuf from locdata.
    The manifest holds metadata for smlm files.

    Parameters
    ----------
    locdata : LocData
        The LocData object.
    return_json_string : bool
        Flag for returning json string

    Returns
    -------
    locan.locan_io.locdata.manifest_pb2.Manifest, str
        The manifest
    """
    manifest = manifest_pb2.Manifest()
    format = manifest_format_from_locdata(locdata)
    file_info = manifest_file_info_from_locdata(locdata)

    # required by specifications
    manifest.format_version = "0.2"
    manifest.formats["smlm-table(binary)-0"].CopyFrom(format)
    manifest.files.append(file_info)

    # recommended by specifications
    manifest.name = locdata.meta.identifier
    manifest.description = locdata.meta.comment
    manifest.tags.append("SMLM")
    manifest.thumbnail = ""
    manifest.sample = ""
    manifest.labeling = ""
    manifest.date = _time_string(time.time())

    # optional by specifications
    manifest.author = ""
    manifest.citation = ""
    manifest.email = ""
    manifest.locdata_meta.CopyFrom(locdata.meta)

    if return_json_string:
        json_string = json_format.MessageToJson(manifest,
                                                preserving_proto_field_name=True,
                                                including_default_value_fields=False)
        json_string = _change_upper_to_lower_keys(json_string)
        return json_string
    else:
        return manifest


def _change_upper_to_lower_keys(json_string: str) -> str:
    """ Switch selected key words in json_string from upper to lower letters."""
    json_string = json_string.replace("BINARY", "binary")
    json_string = json_string.replace("TEXT", "text")
    json_string = json_string.replace("TABLE", "table")
    json_string = json_string.replace("IMAGE", "image")
    json_string = json_string.replace("INT", "int")
    json_string = json_string.replace("UINT", "uint")
    json_string = json_string.replace("FLOAT", "float")
    return json_string


def save_SMLM(locdata, path, manifest=None):
    """
    Save LocData attributes in a SMLM single-molecule
    localization (zip) file with manifest.json (version 0.2).

    In the smlm file format we store metadata as a human-
    readable manifest. The data is stored as byte string.

    Note
    ----
    Only selected LocData attributes are saved.
    These are: 'data', 'columns', 'meta'.

    Parameters
    ----------
    locdata : LocData
        The LocData object to be saved.
    path : str, bytes, os.PathLike, file-like
        File path including file name to save to.
    manifest : locan.locan_io.locdata.manifest_pb2.Manifest
        Protobuf with manifest to use instead of an autogenerated manifest.
    """
    if manifest is None:
        manifest_json = manifest_from_locdata(locdata, return_json_string=True)
    elif isinstance(manifest, str):
        manifest_json = manifest
    elif isinstance(manifest, manifest_pb2.Manifest):
        manifest_json = json_format.MessageToJson(manifest,
                                                  preserving_proto_field_name=True,
                                                  including_default_value_fields=False
                                                  )
        manifest_json = _change_upper_to_lower_keys(manifest_json)
    else:
        raise TypeError("Type of manifest is not correct.")

    byte_string = b"".join([locdata.data[column][i].tobytes()
                            for i in range(len(locdata))
                            for column in locdata.data.columns
                            ])

    # save zip file
    with zipfile.ZipFile(path, 'w', compression=zipfile.ZIP_DEFLATED, allowZip64=True) as zf:
        # write manifest
        zf.writestr("manifest.json", manifest_json)
        # write files
        zf.writestr("table-0.bin", byte_string)


def load_SMLM_manifest(path):
    """
    Read manifest.json (version 0.2) from a SMLM single-molecule localization (zip) file.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a SMLM file to load.

    Returns
    -------
    dictionary
        manifest in json format
    """
    zf = zipfile.ZipFile(path, 'r')
    file_names = zf.namelist()
    if "manifest.json" not in file_names:
        raise Exception('invalid file: no manifest.json found in the smlm file.')
    manifest = json.loads(zf.read("manifest.json"))
    assert manifest['format_version'] == '0.2'
    return manifest


def load_SMLM_header(path):
    """
    Read header (manifest) from a SMLM single-molecule localization file and identify column names.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a SMLM file to load.

    Returns
    -------
    list(str)
        A list of dataset property keys as derived from the SMLM identifiers.
    """
    zf = zipfile.ZipFile(path, 'r')
    file_names = zf.namelist()
    if "manifest.json" not in file_names:
        raise Exception('invalid file: no manifest.json found in the smlm file.')
    manifest = json.loads(zf.read("manifest.json"))
    assert manifest['format_version'] == '0.2'

    locdata_columns_list = []
    for file_info in manifest['files']:
        if file_info['type'] != "table":
            logger.info('ignore file with type: %s', file_info['type'])
        else:
            name = file_info['name']
            logger.info(f'loading table {name}....')
            format_key = file_info['format']
            file_format = manifest['formats'][format_key]
            if name not in file_names:
                logger.error('ERROR: Did not find %s in zip file', file_info['name'])
            else:
                headers = file_format['headers']

                column_keys = convert_property_names(properties=headers, property_mapping=locan.constants.SMLM_KEYS)
                locdata_columns_list.append(column_keys)

    if len(locdata_columns_list) == 1:
        return locdata_columns_list[0]
    else:
        return locdata_columns_list


def load_SMLM_file(path, nrows=None, convert=True):
    """
    Load data from a SMLM single-molecule localization file.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a SMLM file to load.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.
    convert : bool
        If True convert types by applying type specifications in locan.constants.PROPERTY_KEYS.

    Returns
    -------
    LocData, list(LocData)
        A new instance of LocData with all localizations. Returns a list of LocData if multiple tables are found.
    """
    zf = zipfile.ZipFile(path, 'r')
    file_names = zf.namelist()
    if "manifest.json" not in file_names:
        raise Exception('invalid file: no manifest.json found in the smlm file.')
    manifest = json.loads(zf.read("manifest.json"))
    assert manifest['format_version'] == '0.2'

    locdatas = []
    for file_info in manifest['files']:
        if file_info['type'] != "table":
            logger.info('ignore file with type: %s', file_info['type'])
        else:
            name = file_info['name']
            logger.debug(f'start loading {name} ...')
            format_key = file_info['format']
            file_format = manifest['formats'][format_key]
            if file_format['mode'] != 'binary':
                raise Exception(f"format mode {file_format['mode']} not supported.")
            else:
                try:
                    table_file = zf.read(file_info['name'])
                except KeyError:
                    logger.error('ERROR: Did not find %s in zip file', file_info['name'])
                    continue
                else:
                    logger.debug(f'loading {len(table_file)} bytes')
                    headers = file_format['headers']
                    dtype = file_format['dtype']
                    shape = file_format['shape']
                    cols = len(headers)
                    rows = nrows if nrows is not None else int(file_info['rows'])

                    logger.debug('columns: %s', headers)
                    logger.debug('rows: %s, columns: %s', rows, cols)
                    assert len(headers) == len(dtype) == len(shape)

                    rowLen = sum(dtype2length[dtype[i]] for i, header in enumerate(headers))
                    tableDict = {}
                    byteOffset = 0
                    for i, header in enumerate(headers):
                        tableDict[header] = np.ndarray((rows,), buffer=table_file, dtype=dtype[i], offset=byteOffset,
                                                       order='C', strides=(rowLen,))
                        byteOffset += dtype2length[dtype[i]]
                    logger.debug(f'finished loading {name}')

                    dataframe = pd.DataFrame.from_dict(tableDict)

                    column_keys = {}
                    for header in headers:
                        if header in locan.constants.SMLM_KEYS:
                            column_keys[header] = locan.constants.SMLM_KEYS[header]
                        elif header in locan.constants.RAPIDSTORM_KEYS:
                            column_keys[header] = locan.constants.RAPIDSTORM_KEYS[header]
                        else:
                            logger.warning(f'Column {header} is not a Locan property standard.')
                            column_keys[header] = header
                    dataframe.rename(columns=column_keys, inplace=True)

                    if convert:
                        dataframe = convert_property_types(dataframe, types=locan.constants.PROPERTY_KEYS)

                    locdata = LocData.from_dataframe(dataframe=dataframe)

                    # todo: check on taking correct metadata
                    locdata.meta.source = metadata_pb2.EXPERIMENT
                    locdata.meta.state = metadata_pb2.RAW
                    locdata.meta.file.type = metadata_pb2.SMLM
                    locdata.meta.file.path = str(path)

                    for property_ in sorted(list(set(dataframe.columns).intersection(
                            {'position_x', 'position_y', 'position_z'}))):
                        locdata.meta.localization_properties.add(name=property_, unit="nm", type="float")

                    del locdata.meta.history[:]
                    locdata.meta.history.add(name='load_SMLM_file', parameter=f'path={path}, nrows={nrows}')

                    locdatas.append(locdata)

        if len(locdatas) == 1:
            return locdatas[0]
        else:
            return locdatas



import pytest
from locan import IMPORT_NAMES, INSTALL_REQUIRES, EXTRAS_REQUIRE, HAS_DEPENDENCY, needs_package


def test_IMPORT_NAMES():
    package = "numpy"
    assert IMPORT_NAMES.get(package, package) == "numpy"
    package = "scikit-learn"
    assert IMPORT_NAMES.get(package, package) == "sklearn"


def test_dependency_lists():
    assert len(INSTALL_REQUIRES) > 0
    assert len(INSTALL_REQUIRES) == len(set(INSTALL_REQUIRES))
    assert len(EXTRAS_REQUIRE) >= 0
    # print("install_requires:", INSTALL_REQUIRES)
    # print("extras_require:", EXTRAS_REQUIRE)


def test_has_dependency():
    assert "numpy" in HAS_DEPENDENCY
    assert all(HAS_DEPENDENCY)


def test_needs_package():
    assert callable(needs_package("numpy"))


@pytest.mark.skipif(not HAS_DEPENDENCY["pytest"])
@needs_package("pytest")
def function_to_be_decorated():
    """This is documentation for function_to_be_decorated."""
    return True


def test_function_to_be_decorated():
    assert function_to_be_decorated.__doc__ == "This is documentation for function_to_be_decorated."
    assert function_to_be_decorated()



import pytest
import numpy as np

import boost_histogram as bh
from locan.data.aggregate import _is_scalar, _is_single_element, \
    _is_1d_array_of_scalar, _is_1d_array_of_two_scalar,\
    _is_2d_homogeneous_array, _is_2d_inhomogeneous_array, _is_2d_inhomogeneous_array_of_1d_array_of_scalar, \
    _n_bins_to_bin_edges_one_dimension, _bin_size_to_bin_edges_one_dimension, \
    _bin_edges_to_n_bins_one_dimension, _bin_edges_to_n_bins, \
    _bin_edges_to_bin_size_one_dimension, _bin_edges_to_bin_size, \
    _bin_edges_to_bin_centers, _indices_to_bin_centers, \
    _BinsFromBoostHistogramAxis, _BinsFromEdges, _BinsFromNumber, _BinsFromSize
from locan import histogram, Bins

data_scalars = {
    '1': 1,
    '()': ()
}

data_tuples = {
    '((),)': ((),),
    '(1,)': (1,),
    '(1, 2)': (1, 2),
    '(1, 2, 3)': (1, 2, 3),
    '((1, 2),)': ((1, 2),),
    '((1, 2), (1, 2))': ((1, 2), (1, 2)),
    '((1, 2), (1, 2, 3))': ((1, 2), (1, 2, 3)),
    '((1, 2), ((1, 2), (1, 2)))': ((1, 2), ((1, 2), (1, 2))),
    '(1, (1, 2))': (1, (1, 2)),
}

data_lists = {
    '[[]]': [[]],
    '[1]': [1],
    '[1, 2]': [1, 2],
    '[1, 2, 3]': [1, 2, 3],
    '[[1, 2]]': [[1, 2]],
    '[[1, 2], [1, 2]]': [[1, 2], [1, 2]],
    '[[1, 2], [1, 2, 3]]': [[1, 2], [1, 2, 3]],
    '[[1, 2], [[1, 2], [1, 2]]]': [[1, 2], [[1, 2], [1, 2]]],
    '[1, [1, 2]]': [1, [1, 2]],
}

data_ndarrays = {
    'np.array((1))': np.array((1)),
    'np.array((1,))': np.array((1,)),
    'np.array((1, 2))': np.array((1, 2)),
    'np.array([(1, 2)])': np.array([(1, 2)]),
    'np.array([(1, 2)], dtype=object)': np.array([(1, 2)], dtype=object),
    'np.array([(1, 2), (1, 2, 3)], dtype=object)': np.array([(1, 2), (1, 2, 3)], dtype=object)
}

data_all = {**data_scalars, **data_tuples, **data_lists, **data_ndarrays}

expect_is_scalar = ['1', 'np.array((1))']

expect_is_single_element = ['1', '(1,)', '[1]', 'np.array((1))', 'np.array((1,))']

expect_is_1d_array_of_scalar = ['(1, 2)', '(1, 2, 3)',
                                '[1, 2]', '[1, 2, 3]',
                                'np.array((1, 2))',
                                '(1,)', '[1]', 'np.array((1,))']

expect_is_1d_array_of_two_scalar = ['(1, 2)', '[1, 2]', 'np.array((1, 2))']

expect_is_2d_homogeneous_array = ['((1, 2),)', '((1, 2), (1, 2))',
                                  '[[1, 2]]', '[[1, 2], [1, 2]]',
                                  'np.array([(1, 2)])']

expect_is_2d_inhomogeneous_array = ['((1, 2), (1, 2, 3))',
                                    '(1, (1, 2))',
                                    '[[1, 2], [1, 2, 3]]',
                                    '[1, [1, 2]]',
                                    'np.array([(1, 2), (1, 2, 3)], dtype=object)']

expect_is_2d_inhomogeneous_array_of_1d_array_of_scalar = ['((1, 2), (1, 2, 3))',
                                                          '[[1, 2], [1, 2, 3]]',
                                                          'np.array([(1, 2), (1, 2, 3)], dtype=object)']


def test__is_scalar():
    for key, value in data_all.items():
        if key in expect_is_scalar:
            assert _is_scalar(value)
        else:
            assert not _is_scalar(value)


def test__is_single_element():
    for key, value in data_all.items():
        if key in expect_is_single_element:
            assert _is_single_element(value)
        else:
            assert not _is_single_element(value)


def test__is_1d_array_of_scalar():
    for key, value in data_all.items():
        if key in expect_is_1d_array_of_scalar:
            assert _is_1d_array_of_scalar(value)
        else:
            assert not _is_1d_array_of_scalar(value)


def test__is_1d_array_of_two_scalar():
    for key, value in data_all.items():
        if key in expect_is_1d_array_of_two_scalar:
            assert _is_1d_array_of_two_scalar(value)
        else:
            assert not _is_1d_array_of_two_scalar(value)


def test__is_2d_homogeneous_array():
    for key, value in data_all.items():
        if key in expect_is_2d_homogeneous_array:
            assert _is_2d_homogeneous_array(value)
        else:
            assert not _is_2d_homogeneous_array(value)


def test__is_2d_inhomogeneous_array():
    for key, value in data_all.items():
        if key in expect_is_2d_inhomogeneous_array:
            assert _is_2d_inhomogeneous_array(value)
        else:
            assert not _is_2d_inhomogeneous_array(value)


def test__is_2d_inhomogeneous_array_of_1d_array_of_scalar():
    for key, value in data_all.items():
        if key in expect_is_2d_inhomogeneous_array_of_1d_array_of_scalar:
            assert _is_2d_inhomogeneous_array_of_1d_array_of_scalar(value)
        else:
            assert not _is_2d_inhomogeneous_array_of_1d_array_of_scalar(value)


def test__n_bins_to_bin_edges_one_dimension():
    bin_edges = _n_bins_to_bin_edges_one_dimension(10, (10, 20))
    assert bin_edges.shape == (11,)


def test__bin_size_to_bin_edges_one_dimension():
    bin_edges = _bin_size_to_bin_edges_one_dimension(4, (1, 10), extend_range=None)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 5, 9))
    bin_edges = _bin_size_to_bin_edges_one_dimension(4, (1, 10), extend_range=True)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 5, 9, 13))
    bin_edges = _bin_size_to_bin_edges_one_dimension(4, (1, 10), extend_range=False)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 5, 9, 10))

    bin_edges = _bin_size_to_bin_edges_one_dimension(20, (1, 10), extend_range=None)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 10))
    bin_edges = _bin_size_to_bin_edges_one_dimension(20, (1, 10), extend_range=True)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 21))
    bin_edges = _bin_size_to_bin_edges_one_dimension(20, (1, 10), extend_range=False)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 10))

    bin_edges = _bin_size_to_bin_edges_one_dimension((1, 2, 3, 3, 2), (1, 11), extend_range=None)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 2, 4, 7, 10))
    bin_edges = _bin_size_to_bin_edges_one_dimension((1, 2, 3, 3, 2), (1, 11), extend_range=True)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 2, 4, 7, 10, 12))
    bin_edges = _bin_size_to_bin_edges_one_dimension((1, 2, 3, 3, 2), (1, 11), extend_range=False)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 2, 4, 7, 10, 11))

    bin_edges = _bin_size_to_bin_edges_one_dimension((10, 20, 30), (1, 2), extend_range=None)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 2))
    bin_edges = _bin_size_to_bin_edges_one_dimension((10, 20, 30), (1, 2), extend_range=True)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 11))
    bin_edges = _bin_size_to_bin_edges_one_dimension((10, 20, 30), (1, 2), extend_range=False)
    isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 2))

    with pytest.raises(TypeError):
        _bin_size_to_bin_edges_one_dimension(((4,), 2), (0, 10), extend_range=False)


def test__bin_edges_to_n_bins_one_dimension():
    n_bins = _bin_edges_to_n_bins_one_dimension((1, 3, 5))
    assert n_bins == 2
    n_bins = _bin_edges_to_n_bins_one_dimension([1, 2, 4])
    assert n_bins == 2


def test__bin_edges_to_n_bins():
    n_bins = _bin_edges_to_n_bins([1, 3, 5])
    assert n_bins == (2,)
    n_bins = _bin_edges_to_n_bins(([1, 3, 5],))
    assert n_bins == (2,)
    n_bins = _bin_edges_to_n_bins([1, 2, 4])
    assert n_bins == (2,)
    n_bins = _bin_edges_to_n_bins(((1, 3, 5), (1, 2, 4, 5)))
    assert np.array_equal(n_bins, (2, 3))
    n_bins = _bin_edges_to_n_bins([[1, 3, 5], [1, 2, 3, 4]])
    assert np.array_equal(n_bins, (2, 3))
    n_bins = _bin_edges_to_n_bins(np.array([[1, 3, 5], [1, 2, 3]]))
    assert np.array_equal(n_bins, (2, 2))


def test__bin_edges_to_bin_size_one_dimension():
    bin_size = _bin_edges_to_bin_size_one_dimension((1, 3, 5))
    assert bin_size == 2
    bin_size = _bin_edges_to_bin_size_one_dimension([1, 2, 4])
    assert np.array_equal(bin_size, (1, 2))
    bin_size = _bin_edges_to_bin_size_one_dimension([1, 2])
    assert bin_size == 1


def test__bin_edges_to_bin_size():
    bin_size = _bin_edges_to_bin_size([1, 3, 5])
    assert bin_size == (2,)
    bin_size = _bin_edges_to_bin_size(([1, 3, 5],))
    assert bin_size == (2,)
    bin_size = _bin_edges_to_bin_size([1, 2, 4])
    assert np.array_equal(bin_size[0], (1, 2))
    bin_size = _bin_edges_to_bin_size(((1, 3, 5), (1, 2, 4, 5)))
    assert bin_size[0] == 2
    assert np.array_equal(bin_size[1], (1, 2, 1))
    bin_size = _bin_edges_to_bin_size([[1, 3, 5], [1, 2, 3, 4]])
    assert bin_size == (2, 1)
    bin_size = _bin_edges_to_bin_size(np.array([[1, 3, 5], [1, 2, 3]]))
    assert bin_size == (2, 1)


def test__bin_edges_to_bin_centers():
    bin_centers = _bin_edges_to_bin_centers([1, 3, 5])
    assert np.array_equal(bin_centers, ((2, 4),))

    bin_centers = _bin_edges_to_bin_centers(([1, 3, 5],))
    assert np.array_equal(bin_centers, ((2, 4),))

    bin_centers = _bin_edges_to_bin_centers(((1, 3, 5), (1, 2, 4, 6)))
    expected = ((2, 4), (1.5, 3, 5))
    for bc, ex in zip(bin_centers, expected):
        assert np.array_equal(bc, ex)

    bin_edges = np.array([[0, 1, 2, 4, 8, 9], [0, 1, 4, 8]], dtype=object)
    bin_centers = _bin_edges_to_bin_centers(bin_edges)
    expected = (np.array([0.5, 1.5, 3., 6, 8.5]), np.array([0.5, 2.5, 6]))
    for bc, ex in zip(bin_centers, expected):
        assert np.array_equal(bc, ex)


def test__indices_to_bin_centers():
    indices = 2
    bin_edges = np.array([0, 1, 2, 4, 8, 9])
    bin_centers = _indices_to_bin_centers(bin_edges, indices)
    expected = 3
    assert np.array_equal(bin_centers, expected)

    indices = np.array([0, 2, 1])
    bin_edges = np.array([0, 1, 2, 4, 8, 9])
    bin_centers = _indices_to_bin_centers(bin_edges, indices)
    expected = np.array([0.5, 3, 1.5])
    assert np.array_equal(bin_centers, expected)

    indices = np.array([[0, 1], [2, 2], [4, 3]])
    bin_edges = np.array([0, 1, 2, 4, 8, 9])
    bin_centers = _indices_to_bin_centers(bin_edges, indices)
    expected = np.array([[0.5, 1.5], [3, 3], [8.5, 6]])
    assert np.array_equal(bin_centers, expected)

    indices = np.array([[0, 1], [2, 2], [4, 3]])
    bin_edges = np.array([[0, 1, 2, 4, 8, 9], [1, 2, 4, 8, 9]], dtype=object)
    bin_centers = _indices_to_bin_centers(bin_edges, indices)
    expected = np.array([[0.5, 3], [3, 6], [8.5, 8.5]])
    assert np.array_equal(bin_centers, expected)


def test__BinsFromBoostHistogramAxis():
    bhaxis = bh.axis.Regular(5, 0, 10)
    bins = _BinsFromBoostHistogramAxis(bins=bhaxis)
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == ((2, 2, 2, 2, 2),)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3, 5, 7, 9]))

    bhaxes = bh.axis.AxesTuple((bh.axis.Regular(5, 0, 10), bh.axis.Regular(2, 0, 10)))
    bins = _BinsFromBoostHistogramAxis(bins=bhaxes)
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 10.0))
    assert bins.n_bins == (5, 2)
    assert bins.bin_size == ((2.0, 2.0, 2.0, 2.0, 2.0), (5.0, 5.0))
    expected_edges = [np.array([0, 2, 4, 6, 8, 10]), np.array([0, 5, 10])]
    for bin_edges, edges in zip(bins.bin_edges, expected_edges):
        assert np.array_equal(bin_edges, edges)
    expected_centers = [np.array([1, 3, 5, 7, 9]), np.array([2.5, 7.5])]
    for bin_centers, expected_cents in zip(bins.bin_centers, expected_centers):
        assert np.array_equal(bin_centers, expected_cents)


def test__BinsFromEdges():
    bins = _BinsFromEdges(bin_edges=(0, 2, 4))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 4.0),)
    assert bins.n_bins == (2,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4]))

    bins = _BinsFromEdges(bin_edges=((0, 2, 4),))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 4.0),)
    assert bins.n_bins == (2,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4]))

    bins = _BinsFromEdges(bin_edges=(1, 2, 5))
    assert bins.dimension == 1
    assert bins.bin_range == ((1, 5),)
    assert bins.n_bins == (2,)
    assert np.array_equal(bins.bin_size[0], np.array([1, 3]))
    assert np.array_equal(bins.bin_edges[0], np.array([1, 2, 5]))

    bins = _BinsFromEdges(bin_edges=((0, 2, 4), (1, 2, 5)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 4.0), (1.0, 5.0))
    assert bins.n_bins == (2, 2)
    assert bins.bin_size[0] == 2
    assert np.array_equal(bins.bin_size[1], np.array([1, 3]))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4]))
    assert np.array_equal(bins.bin_edges[-1], np.array([1, 2, 5]))

    with pytest.raises(TypeError):
        _BinsFromEdges(bin_edges=())
    with pytest.raises(TypeError):
        _BinsFromEdges(bin_edges=[(0, 2, 4), ((0, 2, 4), (0, 2, 4))])
    with pytest.raises(TypeError):
        _BinsFromEdges(bin_edges=[(0, 2, 4), 2])


def test__BinsFromNumber_():
    bins = _BinsFromNumber(n_bins=5, bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))

    bins = _BinsFromNumber(n_bins=(5,), bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))

    bins = _BinsFromNumber(n_bins=(2, 5), bin_range=((0, 10), (0, 5)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0, 10), (0, 5))
    assert bins.n_bins == (2, 5)
    assert bins.bin_size == (5, 1)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 5, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 2, 3, 4, 5]))

    bins = _BinsFromNumber(n_bins=2, bin_range=((0, 10), (0, 5)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0, 10), (0, 5))
    assert bins.n_bins == (2, 2)
    assert bins.bin_size == (5, 2.5)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 5, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 2.5, 5]))

    bins = _BinsFromNumber(n_bins=(2, 5, 2), bin_range=(0, 10))
    assert bins.dimension == 3
    assert bins.bin_range == ((0, 10), (0, 10), (0, 10))
    assert bins.n_bins == (2, 5, 2)
    assert bins.bin_size == (5, 2, 5)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 5, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_edges[2], np.array([0, 5, 10]))

    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=5, bin_range=1)
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=5, bin_range=(0,))
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=5, bin_range=(1, 2, 3))
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(5, (1, 2)), bin_range=(1, 2, 3))

    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(5,), bin_range=1)
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(5,), bin_range=(0,))
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(5,), bin_range=(1, 2, 3))
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(2,), bin_range=((0, 10), (0, 5)))

    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(2, 5, 2), bin_range=1)
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(2, 5, 2), bin_range=(0,))
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(2, 5, 2), bin_range=(1, 2, 3))

    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(5, (1, 2)), bin_range=(1, 2))


def test__BinsFromSize():
    bins = _BinsFromSize(bin_size=2, bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))

    bins = _BinsFromSize(bin_size=(2,), bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))

    bins = _BinsFromSize(bin_size=((1, 2, 3, 4),), bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (4,)
    assert np.array_equal(bins.bin_size[0], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 1, 3, 6, 10]))

    bins = _BinsFromSize(bin_size=3, bin_range=(0, 10), extend_range=False)
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (4,)
    assert np.array_equal(bins.bin_size[0], (3.0, 3.0, 3.0, 1.0))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 3, 6, 9, 10]))

    bins = _BinsFromSize(bin_size=(5, 1), bin_range=((0, 10), (0, 5)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0, 10), (0, 5))
    assert bins.n_bins == (2, 5)
    assert bins.bin_size == (5, 1)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 5, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 2, 3, 4, 5]))

    bins = _BinsFromSize(bin_size=2, bin_range=((0, 10), (0, 5)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0, 10), (0, 4))
    assert bins.n_bins == (5, 2)
    assert bins.bin_size == (2, 2)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 2, 4]))

    bins = _BinsFromSize(bin_size=(2, 5), bin_range=(0, 10))
    assert bins.dimension == 2
    assert bins.bin_range == ((0, 10), (0, 10))
    assert bins.n_bins == (5, 2)
    assert bins.bin_size == (2, 5)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 5, 10]))

    bins = _BinsFromSize(bin_size=((1, 2, 3, 4), (1, 2, 3, 1)), bin_range=(0, 10))
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 7.0))
    assert bins.n_bins == (4, 4)
    assert np.array_equal(bins.bin_size[0], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_size[1], (1, 2, 3, 1))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 1, 3, 6, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 3, 6, 7]))

    bins = _BinsFromSize(bin_size=((1, 2, 3, 4), (1, 2, 3, 4)), bin_range=((0, 10), (0, 10)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 10.0))
    assert bins.n_bins == (4, 4)
    assert np.array_equal(bins.bin_size[0], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_size[1], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 1, 3, 6, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 3, 6, 10]))

    bins = _BinsFromSize(bin_size=(2, (1, 2, 3, 4)), bin_range=(0, 10))
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 10.0))
    assert bins.n_bins == (5, 4)
    assert bins.bin_size[0] == 2
    assert np.array_equal(bins.bin_size[1], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 3, 6, 10]))

    bins = _BinsFromSize(bin_size=(2, (1, 2, 3, 4)), bin_range=((0, 10), (0, 20)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 10.0))
    assert bins.n_bins == (5, 4)
    assert bins.bin_size[0] == 2
    assert np.array_equal(bins.bin_size[1], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 3, 6, 10]))

    with pytest.raises(TypeError):
        _BinsFromSize(bin_size=5, bin_range=(0,))
    with pytest.raises(TypeError):
        _BinsFromSize(bin_size=5, bin_range=(1, 2, 3))
    with pytest.raises(TypeError):
        _BinsFromSize(bin_size=(2, 2, (1, 2, 3, 4)), bin_range=((0, 10), (0, 20)))
    with pytest.raises(TypeError):
        _BinsFromSize(bin_size=((1, 2, 3, 4), (1, 2, 3, 4), (1, 2, 3, 4)), bin_range=((0, 10), (0, 20)))


def test_Bins():
    bins = Bins(bin_edges=(0, 2, 4))
    assert bins.dimension == 1
    assert bins.bin_range == ((0, 4),)
    assert bins.n_bins == (2,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3]))
    assert bins.is_equally_sized == (True,)

    bins = Bins(n_bins=5, bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3, 5, 7, 9]))
    assert bins.is_equally_sized == (True,)

    bins = Bins(bin_size=2, bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3, 5, 7, 9]))
    assert bins.is_equally_sized == (True,)

    bins = Bins(bin_size=(2, (1, 2, 3)), bin_range=(0, 10))
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 6.0))
    assert bins.n_bins == (5, 3)
    assert np.array_equal(bins.bin_size[0], 2)
    assert np.array_equal(bins.bin_size[1], (1, 2, 3))
    assert bins.is_equally_sized == (True, False)

    bins = Bins(bins=Bins(n_bins=5, bin_range=(0, 10)))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3, 5, 7, 9]))
    assert bins.is_equally_sized == (True,)

    bins = Bins(n_bins=(2, 5), bin_range=(0, 10), labels=["position_x", "position_y"])
    assert bins.labels == ["position_x", "position_y"]
    assert bins.dimension == 2
    assert bins.is_equally_sized == (True, True)
    bins = Bins(bins=Bins(n_bins=5, bin_range=(0, 10)), labels=["position_x"])
    assert bins.labels == ["position_x"]
    assert bins.dimension == 1
    bins = Bins(bins=Bins(n_bins=5, bin_range=(0, 10), labels=["position_x"]))
    assert bins.labels == ["position_x"]
    assert bins.dimension == 1
    with pytest.raises(ValueError):
        Bins(n_bins=5, bin_range=(0, 10), labels=["position_x", "position_y"])


def test_Bins_with_boost_histogram():
    bhaxis = bh.axis.Regular(5, 0, 10)
    bins = Bins(bins=bhaxis)
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == ((2, 2, 2, 2, 2),)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3, 5, 7, 9]))
    assert bins.is_equally_sized == (True,)


def test_Bins_methods():
    bins = Bins(bin_edges=(0, 1, 2, 4))
    assert bins.dimension == 1
    assert bins.bin_range == ((0, 4),)
    assert bins.n_bins == (3,)
    assert np.array_equal(bins.bin_size[0], (1, 1, 2))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 1, 2, 4]))
    assert bins.is_equally_sized == (False,)

    bins = Bins(bin_edges=(0, 1, 2, 4)).equalize_bin_size()
    assert bins.dimension == 1
    assert bins.bin_range == ((0, 4),)
    assert bins.n_bins == (4,)
    assert bins.bin_size == (1,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 1, 2, 3, 4]))
    assert bins.is_equally_sized == (True,)


def test_histogram(locdata_blobs_2d):
    hist = histogram(locdata_blobs_2d, n_bins=10)
    assert hist.labels == ['position_x', 'position_y', 'counts']
    assert hist.data.dtype == 'float64'
    assert hist.data.ndim == 2
    assert np.max(hist.data) == 7

    hist = histogram(locdata_blobs_2d, n_bins=10, bin_range=((500, 1000), (500, 1000)))
    assert hist.data.ndim == 2
    assert np.max(hist.data) == 5
    assert hist.data.shape == (10, 10)

    hist = histogram(locdata_blobs_2d, bin_size=10, loc_properties='position_x')
    assert hist.labels == ['position_x', 'counts']
    assert hist.data.shape == (89,)

    with pytest.raises(ValueError):
        histogram(locdata_blobs_2d, bin_size=10, loc_properties='position_x', bin_range=((500, 1000), (500, 1000)))

    hist = histogram(locdata_blobs_2d, bin_size=10, loc_properties=['position_x'])
    assert hist.labels == ['position_x', 'counts']
    assert hist.data.shape == (89,)

    hist = histogram(locdata_blobs_2d, bin_size=10, loc_properties=['position_x', 'position_y'])
    assert hist.labels == ['position_x', 'position_y', 'counts']
    assert hist.data.shape == (89, 55)

    hist = histogram(locdata_blobs_2d, bin_size=10, loc_properties=['position_x', 'cluster_label'])
    assert hist.labels == ['position_x', 'cluster_label', 'counts']
    assert hist.data.shape == (89, 1)

    with pytest.raises(ValueError):
        histogram(locdata_blobs_2d, bin_size=10, loc_properties='position_z')

    with pytest.raises(ValueError):
        histogram(locdata_blobs_2d, bin_size=10, loc_properties=['position_z'])

    with pytest.raises(ValueError):
        histogram(locdata_blobs_2d, bin_size=10, loc_properties=['position_x', 'position_z'])

    hist = histogram(locdata_blobs_2d, bin_edges=((500, 600, 700, 800, 900, 1000), (500, 600, 700, 800, 900, 1000)))
    assert hist.data.ndim == 2
    assert np.max(hist.data) == 7

    hist = histogram(locdata_blobs_2d, bin_size=10, other_property='position_y')
    assert hist.labels == ['position_x', 'position_y', 'position_y']
    assert hist.data.shape == (89, 55)


def test_histogram_1d(locdata_1d):
    hist = histogram(locdata_1d, n_bins=10)
    assert hist.labels == ['position_x', 'counts']
    assert hist.data.dtype == 'float64'
    assert hist.data.ndim == 1
    assert np.max(hist.data) == 2
    assert hist.data.shape == (10,)

    hist = histogram(locdata_1d, n_bins=5, bin_range=(5, 10))
    assert np.max(hist.data) == 1
    assert hist.data.shape == (5,)

    hist = histogram(locdata_1d,
                     bin_edges=(5, 6, 7, 8, 9, 10))
    assert hist.data.shape == (5,)

    hist = histogram(locdata_1d, n_bins=10, other_property='intensity')
    assert hist.labels == ['position_x', 'intensity']
    assert hist.data.shape == (10,)
    assert np.nanmax(hist.data) == 125


def test_histogram_3d(locdata_blobs_3d):
    hist = histogram(locdata_blobs_3d, n_bins=10)
    assert hist.labels == ['position_x', 'position_y', 'position_z', 'counts']
    assert hist.data.dtype == 'float64'
    assert hist.data.ndim == 3
    assert np.max(hist.data) == 6
    assert hist.data.shape == (10, 10, 10)

    hist = histogram(locdata_blobs_3d, n_bins=10, bin_range=((500, 1000), (500, 1000), (500, 1000)))
    assert np.max(hist.data) == 4
    assert hist.data.shape == (10, 10, 10)

    hist = histogram(locdata_blobs_3d,
                     bin_edges=((500, 600, 700, 800, 900, 1000),
                                (500, 600, 700, 800, 900, 1000),
                                (500, 600, 700, 800, 900, 1000)
                                ))
    assert hist.data.shape == (5, 5, 5)

    hist = histogram(locdata_blobs_3d, n_bins=10, other_property='position_y')
    assert hist.labels == ['position_x', 'position_y', 'position_z', 'position_y']
    assert hist.data.shape == (10, 10, 10)
    assert np.nanmax(hist.data) == 787


def test_histogram_empty(locdata_empty):
    with pytest.raises(TypeError):
        hist = histogram(locdata_empty, n_bins=10)


def test_histogram_single_value(locdata_single_localization_3d):
    hist = histogram(locdata_single_localization_3d, n_bins=3)
    assert hist.data.shape == (3, 3, 3)
    assert np.array_equal(hist.bins.bin_range, [[1, 2], [1, 2], [1, 2]])

    hist = histogram(locdata_single_localization_3d, bin_size=0.2)
    assert hist.data.shape == (5, 5, 5)
    assert np.array_equal(hist.bins.bin_range, [[1, pytest.approx(2)], [1, pytest.approx(2)], [1, pytest.approx(2)]])

    hist = histogram(locdata_single_localization_3d, bin_size=2)
    assert hist.data.shape == (1, 1, 1)
    assert np.array_equal(hist.bins.bin_range, [[1, 2], [1, 2], [1, 2]])


def test_histogram_2d_negative_values(locdata_2d_negative):
    hist = histogram(locdata_2d_negative, n_bins=10)
    assert hist.labels == ['position_x', 'position_y', 'counts']
    assert hist.data.shape == (10, 10)



"""

File input/output for localization data in SMAP files.

"""
import logging

import pandas as pd

from locan.dependencies import HAS_DEPENDENCY, needs_package
from locan.data.locdata import LocData
import locan.constants
from locan.data import metadata_pb2
from locan.locan_io.locdata.utilities import convert_property_types, convert_property_names

if HAS_DEPENDENCY["h5py"]: import h5py


__all__ = ['load_SMAP_header', 'load_SMAP_file', 'save_SMAP_csv']

logger = logging.getLogger(__name__)


@needs_package("h5py")
def _read_SMAP_header(file):
    """
    Identify column names from a SMAP single-molecule localization file.

    Parameters
    ----------
    file : file-like
        File to read.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """
    # list identifiers
    identifiers = list(file["saveloc"]["loc"].keys())
    column_keys = convert_property_names(properties=identifiers, property_mapping=locan.constants.SMAP_KEYS)
    return column_keys


@needs_package("h5py")
def load_SMAP_header(path):
    """
    Identify column names from a SMAP single-molecule localization file.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the rapidSTORM identifiers.
    """
    with h5py.File(path, 'r') as file:
        return _read_SMAP_header(file)


@needs_package("h5py")
def load_SMAP_file(path, nrows=None, convert=True):
    """
    Load data from a SMAP single-molecule localization file.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a Thunderstorm file to load.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.
    convert : bool
        If True convert types by applying type specifications in locan.constants.PROPERTY_KEYS.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with h5py.File(path, 'r') as file:
        columns = _read_SMAP_header(file)

        if file["saveloc"]["loc"]["frame"].shape == (0,):  # empty file
            logger.warning(f'File does not contain any data.')
            locdata = LocData()

        else:  # file not empty
            data = {}
            for key, value in file["saveloc"]["loc"].items():
                if value.shape is not None:
                    if key in locan.constants.SMAP_KEYS:
                        data[locan.constants.SMAP_KEYS[key]] = list(value[0][:nrows])
                    else:
                        data[key] = list(value[0][:nrows])

            dataframe = pd.DataFrame(data)

            if convert:
                dataframe = convert_property_types(dataframe, types=locan.constants.PROPERTY_KEYS)

            locdata = LocData.from_dataframe(dataframe=dataframe)

    locdata.meta.source = metadata_pb2.EXPERIMENT
    locdata.meta.state = metadata_pb2.RAW
    locdata.meta.file.type = metadata_pb2.SMAP
    locdata.meta.file.path = str(path)

    for property_ in sorted(list(set(columns).intersection({'position_x', 'position_y', 'position_z'}))):
        locdata.meta.localization_properties.add(name=property_, unit="nm", type="float")

    del locdata.meta.history[:]
    locdata.meta.history.add(name='load_SMAP_file', parameter='path={}, nrows={}'.format(path, nrows))

    return locdata


def save_SMAP_csv(locdata, path):
    """
    Save LocData to SMAP-readable csv-file.

    In the csv-file file format we store only localization data with SMAP-readable column names.

    Parameters
    ----------
    locdata : LocData
        The LocData object to be saved.
    path : str, bytes, os.PathLike, file-like
        File path including file name to save to.
    """
    # get data from locdata object
    dataframe = locdata.data

    # create reverse mapping to columns
    inv_map = {v: k for k, v in locan.constants.SMAP_KEYS.items()}

    # rename columns
    dataframe = dataframe.rename(index=str, columns=inv_map, inplace=False)
    valid_smap_columns = [key for key in dataframe.columns if key in locan.constants.SMAP_KEYS]

    # write to csv
    dataframe[valid_smap_columns].to_csv(path, float_format='%.10g', index=False)



"""
Compute localization uncertainty.

Localization uncertainty depends on a number of experimental factors including camera and photophysical characteristics
as outlined in [1]_ [2]_. We provide functions to compute an uncertainty estimate from available localization properties.

References
----------
.. [1] K.I. Mortensen, L. S. Churchman, J. A. Spudich, H. Flyvbjerg, Nat. Methods 7 (2010): 377–384.
.. [2] Rieger B., Stallinga S., The lateral and axial localization uncertainty in super-resolution light microscopy.
   Chemphyschem 17;15(4), 2014:664-70. doi: 10.1002/cphc.201300711

"""
import logging

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import warnings

from locan.analysis.analysis_base import _Analysis


__all__ = ['LocalizationUncertaintyFromIntensity']

logger = logging.getLogger(__name__)


##### The algorithms

def _localization_uncertainty_from_intensity(locdata):

    results = {}
    for v in ['x', 'y', 'z']:
        if 'position_' + v in locdata.data.keys() and 'intensity' in locdata.data.keys():
            if 'psf_sigma_' + v in locdata.data.keys():
                results.update(
                    {'uncertainty_' + v: locdata.data['psf_sigma_' + v] / np.sqrt(locdata.data['intensity'])}
                )
            else:
                results.update({'uncertainty_' + v: 1 / np.sqrt(locdata.data['intensity'])})
        else:
            pass

    return pd.DataFrame(results)


##### The specific analysis classes

class LocalizationUncertaintyFromIntensity(_Analysis):
    """
    Compute the localization uncertainty for each localization's spatial coordinate in locdata.

    Uncertainty is computed as Psf_sigma / Sqrt(Intensity) for each spatial dimension.
    If Psf_sigma is not available Uncertainty is 1 / Sqrt(Intensity).

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        The number of localizations per frame or
        the number of localizations per frame normalized to region_measure(hull).
    """
    count = 0

    def __init__(self, meta=None):
        super().__init__(meta=meta)
        self.results = None

    def compute(self, locdata):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data.

        Returns
        -------
        Analysis class
            Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning('Locdata is empty.')
            return self

        self.results = _localization_uncertainty_from_intensity(locdata=locdata)
        return self



import os

import locan
from locan.__main__ import main


def test_version():
    try:
        assert isinstance(locan._version.version, str)
        assert isinstance(locan._version.version_tuple, tuple)
    except AttributeError:
        pass
    assert locan.__version__
    assert locan.__all__
    # print(dir(locan))
    # print(locan.__version__)
    # print(locan.__all__)


def test_entrypoint(capfd):
    main([])
    captured = capfd.readouterr()
    assert captured.out[:47] == "This is the command line entry point for locan."


def test_entrypoint_from_sys(capfd):
    exit_status = os.system('locan')
    captured = capfd.readouterr()
    assert captured.out[:47] == "This is the command line entry point for locan."
    assert exit_status == 0



"""

This module contains tests that need gui interactions and should therefore not be run automatically.

"""
import pytest
from locan import ROOT_DIR, QT_BINDINGS
from locan.gui.io import file_dialog

pytestmark = pytest.mark.gui


def test_file_dialog():
    print(QT_BINDINGS)
    result = file_dialog(directory=ROOT_DIR, message='Select single file')
    # result = file_dialog()
    print(result)
    assert (len(result) == 1)



#!/usr/bin/env python

"""
Run project test suite with pytest.


To run the script::

    test

Try for instance::

    test

See Also
--------
locan.tests.test
"""
import argparse

from locan.tests import test as sc_test


def _add_arguments(parser):
    parser.add_argument("pytest_args", nargs='?', type=str, default=None,
                        help="string with pytest options.")


def main(args=None):
    parser = argparse.ArgumentParser(description='Run test suite.')
    _add_arguments(parser)
    returned_args = parser.parse_args(args)
    sc_test(args=returned_args.pytest_args)


if __name__ == '__main__':
    main()



"""
Utility functions to deal with exemplary datasets.

The data is located in a separate repository `https://github.com/super-resolution/LocanDatasets`.

When calling a function the datasets are expected to reside in a directory specified by the
`locan.constants.DATASETS_DIR` variable.
If the directory does not exist the exemplary files are downloaded from GitHub.
"""
from pathlib import Path

from locan.dependencies import HAS_DEPENDENCY, needs_package
from locan.locan_io.locdata.io_locdata import load_asdf_file
from locan.configuration import DATASETS_DIR

if HAS_DEPENDENCY["requests"]: import requests


__all__ = ['load_npc', 'load_tubulin']


@needs_package("requests")
def load_npc(**kwargs):
    """
    Locdata representing nuclear pore complexes.

    The data was generated by dSTORM [1]_.
    It shows the gp210 protein of the nuclear pore complex labeled with AlexaFluor647.

    References
    ----------
    .. [1] Löschberger A, van de Linde S, Dabauvalle MC, Rieger B, Heilemann M, Krohne G, Sauer M.,
       Super-resolution imaging visualizes the eightfold symmetry of gp210 proteins around the nuclear pore complex
       and resolves the central channel with nanometer resolution.
       J Cell Sci. 2012, 125:570-5, doi: 10.1242/jcs.098822.

    Parameters
    ----------
    kwargs : dict
        Parameters passed to `locan.load_asdf_file()`.

    Returns
    -------
    LocData
    """
    path = Path(DATASETS_DIR) / 'npc_gp210.asdf'
    if not path.exists():
        DATASETS_DIR.mkdir(exist_ok=True)
        url = "https://raw.github.com/super-resolution/LocanDatasets/main/smlm_data/npc_gp210.asdf"
        response = requests.get(url)
        assert response.status_code == requests.codes.ok
        with open(path, 'wb') as file:
            for chunk in response.iter_content(chunk_size=128):
                file.write(chunk)
    locdata = load_asdf_file(path, **kwargs)
    return locdata


@needs_package("requests")
def load_tubulin(**kwargs):
    """
    Locdata representing microtubules.

    The data was generated by dSTORM [1]_.
    It shows alpha-tubulin as part of microtubules within COS-7 cells.
    Tubulin was targeted by primary IgG-antibodies labeled with AlexaFluor647 (2.1 degree of labeling)
    and recorded over 75_000 frames.

    References
    ----------
    .. [1] Dominic A. Helmerich, Gerti Beliu, and Markus Sauer,
       Multiple-Labeled Antibodies Behave Like Single Emitters in Photoswitching Buffer
       ACS Nano 2020, 14, 10, 12629–12641, DOI: 10.1021/acsnano.0c06099

    Parameters
    ----------
    kwargs : dict
        Parameters passed to `locan.load_rapidSTORM_file()`.

    Returns
    -------
    LocData
    """
    path = Path(DATASETS_DIR) / 'tubulin_cos7.asdf'
    if not path.exists():
        DATASETS_DIR.mkdir(exist_ok=True)
        url = "https://raw.github.com/super-resolution/LocanDatasets/main/smlm_data/tubulin_cos7.asdf"
        response = requests.get(url)
        assert response.status_code == requests.codes.ok
        with open(path, 'wb') as file:
            for chunk in response.iter_content(chunk_size=128):
                file.write(chunk)
    locdata = load_asdf_file(path, **kwargs)
    return locdata
